from canonical_o3  import canonical_retriangulation
from readDataFiles import parse_text

from regina import NTriangulation
from snappy import Manifold

import sys

def isometry_signature(iso_sig):
    """
    Compute the isometry signature of a manifold, that is
    the isomorphism signature (from regina) of the canonical
    retriangulation from snappy.
    """

    # Create Manifold from isomorphism signature
    M = Manifold(NTriangulation.fromIsoSig(iso_sig).snapPea())

    # Try canonize and if necessary randomize to get a proto-canonical
    # triangulation with non-flat tetrahedra
    for i in range(10):
        M.canonize()
        
        if has_flat_tetrahedra(M):
            print >> sys.stderr, "Retrying by randomizing"
            M.randomize()
        else:
            retrig = canonical_retriangulation(M)
            return NTriangulation.fromSnapPea(retrig).isoSig()

    raise Exception("Failed after retrying on %s" % iso_sig)

def has_flat_tetrahedra(manifold):
    """
    Does SnapPy manifold have flat tetraehdra.
    After canonize(), solutionType might still be all positively oriented
    even though some are flat.
    """
    return any(
        [ abs(shape['rect'].imag()) < 1e-7
          for shape in manifold.tetrahedra_shapes() ])

def find_isometric(all_iso_sigs):
    """
    Group into a list of lists of isomorphism signatures of
    isometric triangulations.
    """

    # Dictionary: isometry signature |-> isomorphism signatures
    # Each value will be a list of isomorpism signatures of isometric
    # triangulations
    d = {}
    for i, iso_sig in enumerate(all_iso_sigs):
        print >>sys.stderr, "%s -> " % iso_sig,
        # Compute isometry signature
        isometry_sig = isometry_signature(iso_sig)
        print >>sys.stderr, "%s (%d/%d)" % (
            isometry_sig, i+1, len(all_iso_sigs))
        # Add to iso sig to dict
        d.setdefault(isometry_sig,[]).append(iso_sig)

    # Sort all isomorphism signatures
    list_of_list_of_iso_sigs = [ sorted(iso_sigs) + [ isometry_sig ]
                                 for isometry_sig, iso_sigs in d.items() ]
    return sorted(list_of_list_of_iso_sigs)

# Turn a list of list of isomorphism signatures into a list of pairs
# (name, list of isomorphism signatures)
# name is generated from a template that is given two integers
# The first integer is the number of tetrahedra
# The next integer is an index that is per number of tetrahedra
def gen_names(listOfListOfIsoSigs, template):

    # Current index for a particular number of tetrahedra
    # Hard coded to at most 99 tetrahedra
    indicies = [ 0 for i in range(100) ]

    result = []

    # Iterate through input
    for listOfIsoSigs in listOfListOfIsoSigs:
        # Pick first iso sig as representative
        representative = listOfIsoSigs[0]
        # Get the number of tetrahedra
        num_tets = NTriangulation.isoSigComponentSize(representative)
        # Create the name using number of tetrahedra and index
        name = template % (num_tets, indicies[num_tets])
        result.append([name] + listOfIsoSigs)

        # Increment index
        indicies[num_tets] += 1

    return result

def output_list_of_list(list_of_list):
    """
    Output homeomorphic isomorphism signatures on the same line.
    """

    for l in list_of_list:
        print ' '.join(l)

if __name__ == '__main__':
    if not len(sys.argv) == 3:
        print >>sys.stderr, (
            "Usage: python "
            "identifyAndNameIsometricIsomoSigsOfTetrahedralTessellations.py "
            "NAMING_TEMPLATE INPUT_FILE")
        sys.exit(1)

    input_file, template = sys.argv[1:]

    # Read input file
    lines = open(input_file).read()

    # Input file has no groupings, so each line has exactly one entry,
    # get it.
    all_iso_sigs = [ isoSigs.strip()
                     for isoSigs in lines.split('\n')
                     if isoSigs.strip() ]

    # Group by isometry signatures
    list_of_list_of_iso_sigs = find_isometric(all_iso_sigs)

    # Name
    named_list_of_list_of_iso_sigs = gen_names(list_of_list_of_iso_sigs, template)

    # Output
    output_list_of_list(named_list_of_list_of_iso_sigs)

